#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <time.h>
#include "../sversion.h"
#include "../libz/zlib.h"


typedef unsigned char *lpspi_uchar_t;
typedef const unsigned char *lpspi_cuchar_t;
typedef int (PASCAL *lpspi_progresscallback_t)(int nNum, int nDenom, long lData);

#include <pshpack1.h>
typedef struct
{
	unsigned char method[8];    /* k@̎ */
	unsigned long position;     /* t@Cł̈ʒu */
	unsigned long compsize;     /* kꂽTCY */
	unsigned long filesize;     /* ̃t@CTCY */
	time_t timestamp;           /* t@C̍XV */
	char path[200];             /* ΃pX */
	char filename[200];         /* t@Cl[ */
	unsigned long crc;          /* CRC */
} spi_fileinfo_t, *lpspi_fileinfo_t;
#include <poppack.h>

typedef struct
{
	unsigned int f;
	/* --- */
	HANDLE file_h;
	/* --- */
	HLOCAL *mem_r;
	HLOCAL mem_h;
	lpspi_uchar_t mem_p;
	unsigned long mem_len;
	unsigned long mem_pos;
	/* --- */
	unsigned char i_buf[0x4000];
	unsigned char o_buf[0x4000];
} spi_ostream_t, *lpspi_ostream_t;

typedef struct
{
	unsigned int f;
	unsigned long length;
	/* --- */
	HANDLE file_h;
	lpspi_uchar_t mem_p;
	unsigned long mem_pos;
	/* --- */
	unsigned long istop;
	unsigned long ispos;
	unsigned long islen;
	unsigned char isbuf[8192];
	/* --- */
	unsigned long num;
	unsigned long headofs;
	unsigned long infolen;
	lpspi_uchar_t infbufp;
	/* for HELLOWORLD */
	unsigned long infocmp;
	lpspi_uchar_t cmpbufp;
	/* for KIKOKUGAI_DEMO */
	unsigned long nameofs;
	enum
	{
		TYPE_HELLOWORLD,
		TYPE_KIKOKUGAI_DEMO
	} type;
} spi_istream_t, *lpspi_istream_t;

enum
{
	SPI_ERR_SUCCESS = 0,
	SPI_ERR_NOTIMPLEMENTED = -1,
	SPI_ERR_ABORT_BY_USER = 1,
	SPI_ERR_UNKNOWN_FORMAT = 2,
	SPI_ERR_DATA_BROKEN = 3,
	SPI_ERR_SHORT_OF_MEMORY = 4,
	SPI_ERR_MEMORY_ERROR = 5,
	SPI_ERR_FILE_READ = 6,
	SPI_ERR_RESERVED = 7,
	SPI_ERR_INTERNAL = 8
};
typedef int spi_errorcode_t;

static unsigned spi_strlinelen(LPCSTR p)
{
	unsigned l = 0;
	while (*p != 0x0d) p++, l++;
	return l;
}

static unsigned spi_splitpath(LPSTR p, unsigned m)
{
	unsigned r,l;
	for (r = l = 0; l < m; l++)
	{
		if (IsDBCSLeadByte(p[l])) { l++; continue; }
		if (p[l] == ':' || p[l] == '/' || p[l] == '\\')
			r = l + 1;
	}
	return r;
}

static void spi_memcpy(LPVOID d, LPCVOID s, unsigned l)
{
	unsigned p;
	for (p = 0; p < l; p++) ((LPSTR)d)[p] = ((LPCSTR)s)[p];
}

static void spi_memset(LPVOID d, char c, unsigned l)
{
	unsigned p;
	for (p = 0; p < l; p++) ((LPSTR)d)[p] = c;
}


static unsigned long readdword(lpspi_cuchar_t p)
{
	unsigned long ret;
	ret  = ((unsigned long)p[3]) << (3 * 8);
	ret |= ((unsigned long)p[2]) << (2 * 8);
	ret |= ((unsigned)p[1]) << (1 * 8);
	ret |= ((unsigned)p[0]) << (0 * 8);
	return ret;
}
static unsigned readword(lpspi_cuchar_t p)
{
	unsigned ret;
	ret  = ((unsigned)p[1]) << (1 * 8);
	ret |= ((unsigned)p[0]) << (0 * 8);
	return ret;
}

static void writeword(unsigned char *p, unsigned v)
{
	p[0] = (unsigned char)(v >> (8 * 0));
	p[1] = (unsigned char)(v >> (8 * 1));
}
static void writedword(unsigned char *p, unsigned long v)
{
	p[0] = (unsigned char)(v >> (8 * 0));
	p[1] = (unsigned char)(v >> (8 * 1));
	p[2] = (unsigned char)(v >> (8 * 2));
	p[3] = (unsigned char)(v >> (8 * 3));
}


static int spi_istream_seek(lpspi_istream_t psis, unsigned long pos)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		if (psis->istop <= pos && pos <= psis->istop + psis->islen)
		{
			psis->ispos = pos - psis->istop;
			return 1;
		}
		else
		{
			LONG hi = 0;
			psis->ispos = psis->islen = 0;
			psis->istop = SetFilePointer(psis->file_h, pos, &hi, FILE_BEGIN);
			if (psis->istop != pos) return 0;
		}
		return 1;
	case 1:/* memory image */
		if (psis->length <= pos) return 0;
		psis->mem_pos = pos;
		return 1;
	default:
		return 0;
	}
}
static unsigned int spi_istream_read(lpspi_istream_t psis, LPVOID pbuf, unsigned int len)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		{
			unsigned ret = 0;
			while (ret < len)
			{
				if (psis->islen == psis->ispos)
				{
					DWORD rl = 0;
					psis->istop += psis->islen;
					psis->ispos = psis->islen = 0;
					if (!ReadFile(psis->file_h, psis->isbuf, sizeof(psis->isbuf), &rl, NULL)) return 0;
					psis->islen = rl;
					if (rl == 0) break;
				}
				if (psis->islen > psis->ispos)
				{
					unsigned cpl = psis->islen - psis->ispos;
					if (cpl > len - ret) cpl = len - ret;
					spi_memcpy((lpspi_uchar_t)pbuf + ret, psis->isbuf + psis->ispos, cpl);
					psis->ispos += cpl;
					ret += cpl;
				}
			}
			return ret;
		}
	case 1:/* memory image */
		if (len > psis->length - psis->mem_pos) len = psis->length - psis->mem_pos;
		if (len) spi_memcpy(pbuf, psis->mem_p + psis->mem_pos, len);
		psis->mem_pos += len;
		return len;
	default:
		return 0;
	}
}
static void spi_istream_close(lpspi_istream_t psis)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		if (psis->file_h != INVALID_HANDLE_VALUE) CloseHandle(psis->file_h);
		break;
	case 1:/* memory image */
		break;
	default:
		break;
	}
	if (psis->cmpbufp) LocalFree(psis->cmpbufp);
	if (psis->infbufp) LocalFree(psis->infbufp);
}

static int typecheck_helloworld(lpspi_cuchar_t p, unsigned l, lpspi_istream_t psis)
{
	unsigned long ver, num, headofs, infolen, infocmp;
	ver = readdword(p + 0x0);
	if (ver != 1 && ver != 2) return 0;
	num = readdword(p + 0x4);
	if (num == 0) return 0;
	infolen = readdword(p + 0x8);
	if (infolen == 0) return 0;
	infocmp = readdword(p + 0xc);
	if (infocmp == 0) return 0;
	headofs = (ver == 2) ? 0x114 : 0x10;
	if (readword(p + headofs) != 0x9c78) return 0;
	if (psis)
	{
		psis->num = num;
		psis->infolen = infolen;
		psis->infocmp = infocmp;
		psis->headofs = headofs;
		psis->type = TYPE_HELLOWORLD;
		psis->nameofs = 0;
	}
	return 1;
}
static int typecheck_kikokugaitrial(lpspi_cuchar_t p, unsigned l, lpspi_istream_t psis)
{
	unsigned long num, headofs, infolen;
	num = readdword(p + 0x0);
	if (num-- < 2) return 0;
	infolen = num * 0xc + readdword(p + 0x4);
	if (infolen == 0) return 0;
	headofs = 0xc;
	if (num >= 2 && readdword(p + 0x10) != readdword(p + 0x18)) return 0;
	if (num >= 3 && readdword(p + 0x18) + readdword(p + 0x1c) != readdword(p + 0x24)) return 0;
	if (psis)
	{
		psis->num = num;
		psis->infolen = infolen;
		psis->infocmp = 0;
		psis->headofs = headofs;
		psis->type = TYPE_KIKOKUGAI_DEMO;
		psis->nameofs = num * 0xc;
	}
	return 1;
}
static int typecheck(lpspi_cuchar_t p, unsigned l, lpspi_istream_t psis)
{
	if (typecheck_helloworld(p, l, psis)) return 1;
	return typecheck_kikokugaitrial(p, l, psis);
}

voidpf zcalloc (voidpf opaque, unsigned items, unsigned size)
{
	(void)opaque;
	return LocalAlloc(LPTR, items * size);
}
void  zcfree (voidpf opaque, voidpf ptr)
{
	(void)opaque;
	LocalFree(ptr);
}

static spi_errorcode_t spi_istream_open(lpspi_istream_t psis, LPSTR buf, long len, unsigned int flag)
{
	unsigned long hi;
	unsigned char head[512];
	int zerr;
	z_stream zs;
	psis->f = flag;
	psis->infbufp = 0;
	psis->cmpbufp = 0;
	switch (psis->f)
	{
	case 0:/* disk file */
		psis->file_h = CreateFile(buf, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (psis->file_h == INVALID_HANDLE_VALUE) return SPI_ERR_FILE_READ;
		if (len != 0) return SPI_ERR_NOTIMPLEMENTED;
		psis->length = GetFileSize(psis->file_h, &hi);
		/* psis->file_ofs = len */;
		/* --- */
		psis->ispos = 0;
		psis->islen = 0;
		psis->istop = 0;
		break;
	case 1:/* memory image */
		psis->length = len;
		psis->mem_p = buf;
		psis->mem_pos = 0;
		break;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
	spi_memset(head, 0, 512);
	if (0 == spi_istream_read(psis, head, 512)) return SPI_ERR_FILE_READ;
	if (!typecheck(head, 512, psis)) return SPI_ERR_UNKNOWN_FORMAT;
	if (!spi_istream_seek(psis, psis->headofs)) return SPI_ERR_FILE_READ;
	psis->infbufp = LocalAlloc(LPTR, psis->infolen);
	if (!psis->infbufp) return SPI_ERR_SHORT_OF_MEMORY;
	if (psis->type == TYPE_HELLOWORLD)
	{
		psis->cmpbufp = LocalAlloc(LPTR, psis->infocmp);
		if (!psis->cmpbufp) return SPI_ERR_SHORT_OF_MEMORY;
		if (psis->infocmp != spi_istream_read(psis, psis->cmpbufp, psis->infocmp)) return SPI_ERR_FILE_READ;
		zs.zalloc = zcalloc;
		zs.zfree = zcfree;
		zs.opaque = 0;
		zs.next_in = psis->cmpbufp;
		zs.avail_in = psis->infocmp;
		zs.next_out = psis->infbufp;
		zs.avail_out = psis->infolen;
		zerr = inflateInit(&zs);
		if (zerr == Z_OK)
		{
			zerr = inflate(&zs, Z_FINISH);
			LocalFree(psis->cmpbufp); psis->cmpbufp = 0;
		}
		inflateEnd(&zs);
		if (zerr == Z_MEM_ERROR) return SPI_ERR_SHORT_OF_MEMORY;
		if (zerr != Z_STREAM_END) return SPI_ERR_FILE_READ;
	}
	else /*if (psis->type == TYPE_KIKOKUGAI_DEMO)*/
	{
		if (psis->infolen != spi_istream_read(psis, psis->infbufp, psis->infolen)) return SPI_ERR_FILE_READ;
	}
	return SPI_ERR_SUCCESS;
}

static const char method_raw[8] = "raw\0\0\0\0\0";
static const char method_deflate[8] = "deflate\0";
static const char method_unknown[8] = "unknown\0";

static spi_errorcode_t getinfo(lpspi_istream_t psis, lpspi_fileinfo_t pinfo, unsigned long pos, unsigned *pinfolen)
{
	unsigned long pathlen, nameofs, fnameofs;
	if (psis->type == TYPE_HELLOWORLD)
	{
		pathlen = readdword(psis->infbufp + pos);
		if (pinfolen) *pinfolen = 4 + pathlen + 0x14;
		nameofs = pos + 4;

		pinfo->crc       = readdword(psis->infbufp + pos + 4 + pathlen + 0x00) + psis->headofs + psis->infocmp;
		pinfo->filesize  = readdword(psis->infbufp + pos + 4 + pathlen + 0x08);
		pinfo->timestamp = readdword(psis->infbufp + pos + 4 + pathlen + 0x0c);
		pinfo->compsize  = readdword(psis->infbufp + pos + 4 + pathlen + 0x10);
		if (pinfo->timestamp == 0 || pinfo->filesize == 0)
		{
			pinfo->compsize = pinfo->filesize;
			spi_memcpy(pinfo->method, method_raw, 8);
		}
		else if (pinfo->timestamp == 1)
			spi_memcpy(pinfo->method, method_deflate, 8);
		else
			spi_memcpy(pinfo->method, method_unknown, 8);
	}
	else /*if (psis->type == TYPE_KIKOKUGAI_DEMO)*/
	{
		psis->nameofs += spi_strlinelen(psis->infbufp + psis->nameofs) + 1;
		pathlen = spi_strlinelen(psis->infbufp + psis->nameofs);
		if (pinfolen) *pinfolen = 0xc;
		nameofs = psis->nameofs;

		pinfo->crc      = readdword(psis->infbufp + pos + 0) + psis->headofs + psis->infolen;
		pinfo->filesize = readdword(psis->infbufp + pos + 4);
		pinfo->compsize = pinfo->filesize;
		pinfo->timestamp = 0;
		spi_memcpy(pinfo->method, method_unknown, 8);
	}
	fnameofs = spi_splitpath(psis->infbufp + nameofs, pathlen);
	spi_memcpy(pinfo->path, psis->infbufp + nameofs, fnameofs);
	pinfo->path[fnameofs] = 0;
	spi_memcpy(pinfo->filename, psis->infbufp + nameofs + fnameofs, pathlen - fnameofs);
	pinfo->filename[pathlen - fnameofs] = 0;
	pinfo->position = pos;
	return SPI_ERR_SUCCESS;
}

static spi_errorcode_t spi_ostream_open(lpspi_ostream_t *ppo, LPSTR buf, unsigned long len, unsigned int flag)
{
	lpspi_ostream_t po;
	*ppo = po = LocalAlloc(LPTR, sizeof(spi_ostream_t));
	if (po == NULL) return SPI_ERR_SHORT_OF_MEMORY;
	po->f = flag >> 8;
	switch (po->f)
	{
	case 0:/* disk file */
		po->file_h = CreateFile(buf, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (po->file_h == INVALID_HANDLE_VALUE) return SPI_ERR_FILE_READ;
		break;
	case 1:/* memory image */
		po->mem_r = (HLOCAL *)buf;
		*po->mem_r = 0;
		po->mem_p = 0;
		po->mem_h = LocalAlloc(LMEM_MOVEABLE, len ? len : 1);
		if (po->mem_h == NULL) return SPI_ERR_SHORT_OF_MEMORY;
		po->mem_p = (lpspi_uchar_t)LocalLock(po->mem_h);
		if (po->mem_p == NULL) return SPI_ERR_MEMORY_ERROR;
		po->mem_len = len;
		po->mem_pos = 0;
		break;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
	return SPI_ERR_SUCCESS;
}
static spi_errorcode_t spi_ostream_write(lpspi_ostream_t po, LPSTR buf, unsigned long len)
{
	switch (po->f)
	{
	case 0:/* disk file */
		{
			DWORD r = 0;
			if (!WriteFile(po->file_h, buf, len, &r, NULL)) return SPI_ERR_FILE_READ;
		}
		return SPI_ERR_SUCCESS;
	case 1:/* memory image */
		if (len > po->mem_len - po->mem_pos) len = po->mem_len - po->mem_pos;
		if (len) spi_memcpy(po->mem_p + po->mem_pos, buf, len);
		po->mem_pos += len;
		return SPI_ERR_SUCCESS;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
}

static void spi_ostream_flush(lpspi_ostream_t po)
{
	switch (po->f)
	{
	case 0:/* disk file */
		break;
	case 1:/* memory image */
		LocalUnlock(po->mem_h);
		*po->mem_r = po->mem_h;
		po->mem_p = NULL;
		po->mem_h = NULL;
		break;
	}
}
static void spi_ostream_close(lpspi_ostream_t po)
{
	if (po)
	{
		switch (po->f)
		{
			case 0:/* disk file */
				if (po->file_h != INVALID_HANDLE_VALUE) CloseHandle(po->file_h);
				break;
			case 1:/* memory image */
				if (po->mem_p != NULL) LocalUnlock(po->mem_h);
				if (po->mem_h != NULL) LocalFree(po->mem_h);
				break;
		}
		LocalFree(po);
	}
}

SUSIEPLUGIN_INFO_HWPAK(spi_info)
int PASCAL GetPluginInfo(int infono, LPSTR buf, int buflen)
{
	int l;
	if (((unsigned)infono) >= (sizeof(spi_info) / sizeof(const char *))) return 0;
	l = lstrlen(spi_info[infono]) + 1;
	if (l > buflen) l = buflen;
	spi_memcpy(buf, spi_info[infono], l);
	return l;
}

int PASCAL IsSupported(LPSTR filename,DWORD dw)
{
	unsigned char head[512];
	int ret = 0;
	LONG hi = 0;
	do
	{
		lpspi_uchar_t pdata;
		if (HIWORD(((DWORD)dw)))
		{
			pdata = (lpspi_uchar_t)dw;
		}
		else
		{
			HFILE h;
			h = (HFILE)LOWORD(((DWORD)dw));
			if (0 != SetFilePointer((HANDLE)h, 0, &hi, FILE_BEGIN)) break;
			spi_memset(head, 0, 512);
			if (0 == ReadFile((HANDLE)h, head, 512, &hi, NULL)) break;
			pdata = head;
		}
		ret = typecheck(pdata, 512, 0) != 0;
	} while (0);
	return ret;
}

spi_errorcode_t PASCAL GetArchiveInfo(LPSTR buf, long len, unsigned int flag, HLOCAL *lphInf)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	spi_istream_t sis;
	HLOCAL hinfo = NULL;
	lpspi_fileinfo_t pinfo = NULL;
	/* if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED; */
	do
	{
		unsigned pos, idx, infolen;
		e = spi_istream_open(&sis, buf, len, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		pos = 0;
		idx = 0;
		hinfo = LocalAlloc(LHND, sis.num * sizeof(spi_fileinfo_t) + 1);
		if (hinfo == NULL) { e = SPI_ERR_SHORT_OF_MEMORY; break;}
		pinfo = (lpspi_fileinfo_t)LocalLock(hinfo);
		if (pinfo == NULL) { e = SPI_ERR_MEMORY_ERROR; break;}
		while (idx < sis.num)
		{
			e = getinfo(&sis, pinfo, pos, &infolen);
			if (e != SPI_ERR_SUCCESS) break;
			pinfo->crc = 0;
			pinfo->timestamp = 0;
			pinfo++;
			idx++;
			pos += infolen;
		}
		if (e != SPI_ERR_SUCCESS) break;
		LocalUnlock(hinfo);
		*lphInf = hinfo;
		pinfo = NULL; hinfo = NULL;
	} while (0);
	if (pinfo != NULL) LocalUnlock(hinfo);
	if (hinfo != NULL) LocalFree(hinfo);
	spi_istream_close(&sis);
	return e;
}

spi_errorcode_t PASCAL GetFileInfo(LPSTR buf, long len, LPSTR filename, unsigned int flag, lpspi_fileinfo_t lpInfo)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	spi_istream_t sis;
	/* if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED; */
	do
	{
		unsigned pos, idx, infolen;
		unsigned char fn[200+200];
		e = spi_istream_open(&sis, buf, len, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		pos = 0;
		idx = 0;
		while (idx < sis.num)
		{
			e = getinfo(&sis, lpInfo, pos, &infolen);
			if (e != SPI_ERR_SUCCESS) break;
			lstrcpy(fn, lpInfo->path);
			lstrcat(fn, lpInfo->filename);
			if ((((flag & 0x80)?lstrcmpi:lstrcmp)(fn, filename)) == 0) break;
			idx++;
			pos += infolen;
		}
		if (idx == sis.num) e = SPI_ERR_INTERNAL;
		if (e != SPI_ERR_SUCCESS) break;
		lpInfo->crc = 0;
		lpInfo->timestamp = 0;
	} while (0);
	spi_istream_close(&sis);
	return e;
}

spi_errorcode_t PASCAL GetFile(LPSTR src, long len, LPSTR dest, unsigned int flag, FARPROC prgresisCallback, long lData)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	lpspi_ostream_t po = 0;
	spi_istream_t sis;
	spi_fileinfo_t info;
	if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED;
	do
	{
		unsigned long p, r, l, ctype, pcur, pmax;
		int zerr;
		z_stream zs;
		e = spi_istream_open(&sis, src, 0, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		e = getinfo(&sis, &info, len, 0);
		if (e != SPI_ERR_SUCCESS) break;
		r = info.filesize;
		ctype = info.timestamp;
		e = spi_ostream_open(&po, dest, r, flag & 0x700);
		if (e != SPI_ERR_SUCCESS) break;
		if (!spi_istream_seek(&sis, info.crc)) { e = SPI_ERR_FILE_READ; break; }
		p = 0;
		pcur = ~0;
		pmax = (r + ((1 >> 16) - 1)) >> 16;

		switch (ctype)
		{
		case 1:
			zs.zalloc = zcalloc;
			zs.zfree = zcfree;
			zs.opaque = 0;
			zs.next_in = po->i_buf;
			zs.avail_in = spi_istream_read(&sis, po->i_buf, sizeof(po->i_buf));
			zs.next_out = po->o_buf;
			zs.avail_out = sizeof(po->o_buf);
			zerr = inflateInit(&zs);
			if (zerr != Z_OK)
			{
				e = (zerr == Z_MEM_ERROR) ? SPI_ERR_SHORT_OF_MEMORY : SPI_ERR_FILE_READ;
				break;
			}
			while (p < r)
			{
				if (prgresisCallback && pcur != (p >> 16))
				{
					pcur = (p >> 16);
					if (((lpspi_progresscallback_t)prgresisCallback)(pcur, pmax, lData))
					{
						e = SPI_ERR_ABORT_BY_USER;
						break;
					}
				}
				zerr = inflate(&zs, Z_SYNC_FLUSH);
				if (zerr < 0)
				{
					e = SPI_ERR_FILE_READ;
					break;
				}
				l = sizeof(po->o_buf) - zs.avail_out;
				if (l)
				{
					e = spi_ostream_write(po, po->o_buf, l);
					if (e != SPI_ERR_SUCCESS) break;
					zs.next_out = po->o_buf;
					zs.avail_out = sizeof(po->o_buf);
					p += l;
				}
				if (zerr == Z_STREAM_END) break;
				if (0 == zs.avail_in)
				{
					zs.next_in = po->i_buf;
					zs.avail_in = spi_istream_read(&sis, po->i_buf, sizeof(po->i_buf));
				}
			}
			inflateEnd(&zs);
			break;
		case 0:
			while (p < r)
			{
				if (prgresisCallback && pcur != (p >> 16))
				{
					pcur = (p >> 16);
					if (((lpspi_progresscallback_t)prgresisCallback)(pcur, pmax, lData))
					{
						e = SPI_ERR_ABORT_BY_USER;
						break;
					}
				}
				l = spi_istream_read(&sis, po->i_buf, ((r - p) > sizeof(po->i_buf)) ? sizeof(po->i_buf) : (r - p));
				if (l == 0) { e = SPI_ERR_FILE_READ; break; }
				e = spi_ostream_write(po, po->i_buf, l);
				if (e != SPI_ERR_SUCCESS) break;
				p += l;
			}
			break;
		default:
			e = SPI_ERR_UNKNOWN_FORMAT;
			break;
		}

		if (e != SPI_ERR_SUCCESS) break;
		if (prgresisCallback)
		{
			if (((lpspi_progresscallback_t)prgresisCallback)(pmax, pmax, lData))
			{
				e = SPI_ERR_ABORT_BY_USER;
				break;
			}
		}
		spi_ostream_flush(po);
	} while (0);
	spi_ostream_close(po);
	spi_istream_close(&sis);
	return e;
}

#if defined(_MSC_VER) && !defined(_DEBUG)
#pragma comment(linker, "/STUB:../../stub.exe")
#pragma comment(linker, "/OPT:NOWIN98")
#pragma comment(linker, "/MERGE:.rdata=.text")
#pragma comment(linker, "/ENTRY:DllMain")
#endif
BOOL WINAPI DllMain(HINSTANCE hInstDLL, DWORD dwReason, LPVOID lpReserved)
{
	if (dwReason == DLL_PROCESS_ATTACH)
	{
		DisableThreadLibraryCalls(hInstDLL);
	}
	return 1;
}
