#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <time.h>
#include "../sversion.h"

typedef unsigned char *lpspi_uchar_t;
typedef const unsigned char *lpspi_cuchar_t;
typedef int (PASCAL *lpspi_progresscallback_t)(int nNum, int nDenom, long lData);

#include <pshpack1.h>
typedef struct
{
	unsigned char method[8];    /* k@̎ */
	unsigned long position;     /* t@Cł̈ʒu */
	unsigned long compsize;     /* kꂽTCY */
	unsigned long filesize;     /* ̃t@CTCY */
	time_t timestamp;           /* t@C̍XV */
	char path[200];             /* ΃pX */
	char filename[200];         /* t@Cl[ */
	unsigned long crc;          /* CRC */
} spi_fileinfo_t, *lpspi_fileinfo_t;
#include <poppack.h>

typedef struct
{
	unsigned int f;
	/* --- */
	HANDLE file_h;
	/* --- */
	HLOCAL *mem_r;
	HLOCAL mem_h;
	lpspi_uchar_t mem_p;
	unsigned long mem_len;
	unsigned long mem_pos;
	/* --- */
	signed long ulaw_table[256];
	unsigned char i_buf[0x400];
	unsigned char o_buf[0x400 * 2];
	unsigned char frame_table[0x10000 * 2];
} spi_ostream_t, *lpspi_ostream_t;

typedef struct
{
	unsigned int f;
	unsigned long length;
	/* --- */
	HANDLE file_h;
	lpspi_uchar_t mem_p;
	unsigned long mem_pos;
	/* --- */
	unsigned long istop;
	unsigned long ispos;
	unsigned long islen;
	unsigned char isbuf[8192];
	/* --- */
	unsigned long num;
	unsigned long srate;
	unsigned long headofs;
} spi_istream_t, *lpspi_istream_t;

enum
{
	SPI_ERR_SUCCESS = 0,
	SPI_ERR_NOTIMPLEMENTED = -1,
	SPI_ERR_ABORT_BY_USER = 1,
	SPI_ERR_UNKNOWN_FORMAT = 2,
	SPI_ERR_DATA_BROKEN = 3,
	SPI_ERR_SHORT_OF_MEMORY = 4,
	SPI_ERR_MEMORY_ERROR = 5,
	SPI_ERR_FILE_READ = 6,
	SPI_ERR_RESERVED = 7,
	SPI_ERR_INTERNAL = 8
};
typedef int spi_errorcode_t;

static void spi_memcpy(LPVOID d, LPCVOID s, unsigned l)
{
	unsigned p;
	for (p = 0; p < l; p++) ((LPSTR)d)[p] = ((LPCSTR)s)[p];
}

static void spi_memset(LPVOID d, char c, unsigned l)
{
	unsigned p;
	for (p = 0; p < l; p++) ((LPSTR)d)[p] = c;
}


static unsigned long readdword(lpspi_cuchar_t p)
{
	unsigned long ret;
	ret  = ((unsigned long)p[3]) << (3 * 8);
	ret |= ((unsigned long)p[2]) << (2 * 8);
	ret |= ((unsigned)p[1]) << (1 * 8);
	ret |= ((unsigned)p[0]) << (0 * 8);
	return ret;
}
static unsigned readword(lpspi_cuchar_t p)
{
	unsigned ret;
	ret  = ((unsigned)p[1]) << (1 * 8);
	ret |= ((unsigned)p[0]) << (0 * 8);
	return ret;
}

static void writeword(unsigned char *p, unsigned v)
{
	p[0] = (unsigned char)(v >> (8 * 0));
	p[1] = (unsigned char)(v >> (8 * 1));
}
static void writedword(unsigned char *p, unsigned long v)
{
	p[0] = (unsigned char)(v >> (8 * 0));
	p[1] = (unsigned char)(v >> (8 * 1));
	p[2] = (unsigned char)(v >> (8 * 2));
	p[3] = (unsigned char)(v >> (8 * 3));
}


static int spi_istream_seek(lpspi_istream_t psis, unsigned long pos)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		if (psis->istop <= pos && pos <= psis->istop + psis->islen)
		{
			psis->ispos = pos - psis->istop;
			return 1;
		}
		else
		{
			LONG hi = 0;
			psis->ispos = psis->islen = 0;
			psis->istop = SetFilePointer(psis->file_h, pos, &hi, FILE_BEGIN);
			if (psis->istop != pos) return 0;
		}
		return 1;
	case 1:/* memory image */
		if (psis->length <= pos) return 0;
		psis->mem_pos = pos;
		return 1;
	default:
		return 0;
	}
}
static unsigned int spi_istream_read(lpspi_istream_t psis, LPVOID pbuf, unsigned int len)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		{
			unsigned ret = 0;
			while (ret < len)
			{
				if (psis->islen == psis->ispos)
				{
					DWORD rl = 0;
					psis->istop += psis->islen;
					psis->ispos = psis->islen = 0;
					if (!ReadFile(psis->file_h, psis->isbuf, sizeof(psis->isbuf), &rl, NULL)) return 0;
					psis->islen = rl;
					if (rl == 0) break;
				}
				if (psis->islen > psis->ispos)
				{
					unsigned cpl = psis->islen - psis->ispos;
					if (cpl > len - ret) cpl = len - ret;
					spi_memcpy((lpspi_uchar_t)pbuf + ret, psis->isbuf + psis->ispos, cpl);
					psis->ispos += cpl;
					ret += cpl;
				}
			}
			return ret;
		}
	case 1:/* memory image */
		if (len > psis->length - psis->mem_pos) len = psis->length - psis->mem_pos;
		if (len) spi_memcpy(pbuf, psis->mem_p + psis->mem_pos, len);
		psis->mem_pos += len;
		return len;
	default:
		return 0;
	}
}
static void spi_istream_close(lpspi_istream_t psis)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		if (psis->file_h != INVALID_HANDLE_VALUE) CloseHandle(psis->file_h);
		break;
	case 1:/* memory image */
		break;
	default:
		break;
	}
}

static int typecheck(lpspi_cuchar_t p, unsigned l, lpspi_istream_t psis)
{
	unsigned long num, srate;
	if (readdword(p + 0x00) != 0x50454f4bUL) return 0;
	if (readword(p + 0x04) != 0x4341) return 0;
	num = readdword(p + 0x10);
	srate = readdword(p + 0x18);
	if (srate == 0) srate = 22050;
	if (psis)
	{
		psis->num = num;
		psis->srate = srate;
		psis->headofs = 32;
	}
	return 1;
}

static spi_errorcode_t spi_istream_open(lpspi_istream_t psis, LPSTR buf, long len, unsigned int flag)
{
	unsigned long hi;
	unsigned char head[32];
	psis->f = flag;
	switch (psis->f)
	{
	case 0:/* disk file */
		psis->file_h = CreateFile(buf, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (psis->file_h == INVALID_HANDLE_VALUE) return SPI_ERR_FILE_READ;
		if (len != 0) return SPI_ERR_NOTIMPLEMENTED;
		psis->length = GetFileSize(psis->file_h, &hi);
		/* psis->file_ofs = len */;
		/* --- */
		psis->ispos = 0;
		psis->islen = 0;
		psis->istop = 0;
		break;
	case 1:/* memory image */
		psis->length = len;
		psis->mem_p = buf;
		psis->mem_pos = 0;
		break;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
	if (32 != spi_istream_read(psis, head, 32)) return SPI_ERR_FILE_READ;
	if (!typecheck(head, 32, psis)) return SPI_ERR_UNKNOWN_FORMAT;
	return SPI_ERR_SUCCESS;
}

static const unsigned char wavhead_init[0x28] = {
	0x52,0x49,0x46,0x46, 0x00,0x00,0x00,0x00, 0x57,0x41,0x56,0x45, 0x66,0x6d,0x74,0x20,
	0x10,0x00,0x00,0x00, 0x01,0x00,0x01,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,
	0x02,0x00,0x10,0x00, 0x64,0x61,0x74,0x61,
};
static const char method_koepac[8] = "koepac\0\0";
static const char form_wav[] = "%08d.wav";

static spi_errorcode_t getinfo(lpspi_istream_t psis, lpspi_fileinfo_t pinfo, unsigned long pos, unsigned idx)
{
	unsigned char hbuf[16];
	if (!spi_istream_seek(psis, pos)) return SPI_ERR_FILE_READ;
	if (16 != spi_istream_read(psis, hbuf, 16)) return SPI_ERR_FILE_READ;

	spi_memcpy(pinfo->method, method_koepac, 8);
	pinfo->position = pos;
	if (idx + 1 == psis->num)
		pinfo->compsize = psis->length - readdword(hbuf + 4);
	else
		pinfo->compsize = readdword(hbuf + 4 + 8) - readdword(hbuf + 4);
	pinfo->filesize = 0x2c + readword(hbuf + 2) * 0x400 * 2;
	pinfo->timestamp = readword(hbuf + 2);
	pinfo->path[0] = 0;
	wsprintf(pinfo->filename, form_wav, readword(hbuf + 0));
	pinfo->crc = readdword(hbuf+ 0x4);

	return SPI_ERR_SUCCESS;
}
static spi_errorcode_t spi_ostream_open(lpspi_ostream_t *ppo, LPSTR buf, unsigned long len, unsigned int flag)
{
	lpspi_ostream_t po;
	*ppo = po = LocalAlloc(LPTR, sizeof(spi_ostream_t));
	if (po == NULL) return SPI_ERR_SHORT_OF_MEMORY;
	po->f = flag >> 8;
	switch (po->f)
	{
	case 0:/* disk file */
		po->file_h = CreateFile(buf, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (po->file_h == INVALID_HANDLE_VALUE) return SPI_ERR_FILE_READ;
		break;
	case 1:/* memory image */
		po->mem_r = (HLOCAL *)buf;
		*po->mem_r = 0;
		po->mem_p = 0;
		po->mem_h = LocalAlloc(LMEM_MOVEABLE, len);
		if (po->mem_h == NULL) return SPI_ERR_SHORT_OF_MEMORY;
		po->mem_p = (lpspi_uchar_t)LocalLock(po->mem_h);
		if (po->mem_p == NULL) return SPI_ERR_MEMORY_ERROR;
		po->mem_len = len;
		po->mem_pos = 0;
		break;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
	return SPI_ERR_SUCCESS;
}
static spi_errorcode_t spi_ostream_write(lpspi_ostream_t po, LPSTR buf, unsigned long len)
{
	switch (po->f)
	{
	case 0:/* disk file */
		{
			DWORD r = 0;
			if (!WriteFile(po->file_h, buf, len, &r, NULL)) return SPI_ERR_FILE_READ;
		}
		return SPI_ERR_SUCCESS;
	case 1:/* memory image */
		if (len > po->mem_len - po->mem_pos) len = po->mem_len - po->mem_pos;
		if (len) spi_memcpy(po->mem_p + po->mem_pos, buf, len);
		po->mem_pos += len;
		return SPI_ERR_SUCCESS;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
}

static void spi_ostream_flush(lpspi_ostream_t po)
{
	switch (po->f)
	{
	case 0:/* disk file */
		break;
	case 1:/* memory image */
		LocalUnlock(po->mem_h);
		*po->mem_r = po->mem_h;
		po->mem_p = NULL;
		po->mem_h = NULL;
		break;
	}
}
static void spi_ostream_close(lpspi_ostream_t po)
{
	if (po)
	{
		switch (po->f)
		{
			case 0:/* disk file */
				if (po->file_h != INVALID_HANDLE_VALUE) CloseHandle(po->file_h);
				break;
			case 1:/* memory image */
				if (po->mem_p != NULL) LocalUnlock(po->mem_h);
				if (po->mem_h != NULL) LocalFree(po->mem_h);
				break;
		}
		LocalFree(po);
	}
}

SUSIEPLUGIN_INFO_KOE(spi_info)
int PASCAL GetPluginInfo(int infono, LPSTR buf, int buflen)
{
	int l;
	if (((unsigned)infono) >= (sizeof(spi_info) / sizeof(const char *))) return 0;
	l = lstrlen(spi_info[infono]) + 1;
	if (l > buflen) l = buflen;
	spi_memcpy(buf, spi_info[infono], l);
	return l;
}

int PASCAL IsSupported(LPSTR filename,DWORD dw)
{
	unsigned char head[32];
	int ret = 0;
	LONG hi = 0;
	do
	{
		lpspi_uchar_t pdata;
		if (HIWORD(((DWORD)dw)))
		{
			pdata = (lpspi_uchar_t)dw;
		}
		else
		{
			HFILE h;
			h = (HFILE)LOWORD(((DWORD)dw));
			if (0 != SetFilePointer((HANDLE)h, 0, &hi, FILE_BEGIN)) break;
			ReadFile((HANDLE)h, head, 32, &hi, NULL);
			pdata = head;
		}
		ret = typecheck(pdata, 32, 0) != 0;
	} while (0);
	return ret;
}

spi_errorcode_t PASCAL GetArchiveInfo(LPSTR buf, long len, unsigned int flag, HLOCAL *lphInf)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	spi_istream_t sis;
	HLOCAL hinfo = NULL;
	lpspi_fileinfo_t pinfo = NULL;
	/* if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED; */
	do
	{
		unsigned pos, idx;
		e = spi_istream_open(&sis, buf, len, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		pos = sis.headofs;
		idx = 0;
		hinfo = LocalAlloc(LHND, sis.num * sizeof(spi_fileinfo_t) + 1);
		if (hinfo == NULL) { e = SPI_ERR_SHORT_OF_MEMORY; break;}
		pinfo = (lpspi_fileinfo_t)LocalLock(hinfo);
		if (pinfo == NULL) { e = SPI_ERR_MEMORY_ERROR; break;}
		while (idx < sis.num)
		{
			e = getinfo(&sis, pinfo, pos, idx);
			if (e != SPI_ERR_SUCCESS) break;
			pinfo->crc = 0;
			pinfo->timestamp = 0;
			pinfo++;
			idx++;
			pos += 8;
		}
		if (e != SPI_ERR_SUCCESS) break;
		LocalUnlock(hinfo);
		*lphInf = hinfo;
		pinfo = NULL; hinfo = NULL;
	} while (0);
	if (pinfo != NULL) LocalUnlock(hinfo);
	if (hinfo != NULL) LocalFree(hinfo);
	spi_istream_close(&sis);
	return e;
}

spi_errorcode_t PASCAL GetFileInfo(LPSTR buf, long len, LPSTR filename, unsigned int flag, lpspi_fileinfo_t lpInfo)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	spi_istream_t sis;
	/* if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED; */
	do
	{
		unsigned pos, idx;
		unsigned char fn[200+200];
		e = spi_istream_open(&sis, buf, len, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		pos = sis.headofs;
		idx = 0;
		while (idx < sis.num)
		{
			e = getinfo(&sis, lpInfo, pos, idx);
			if (e != SPI_ERR_SUCCESS) break;
			lstrcpy(fn, lpInfo->path);
			lstrcat(fn, lpInfo->filename);
			if ((((flag & 0x80)?lstrcmpi:lstrcmp)(fn, filename)) == 0) break;
			idx++;
			pos += 8;
		}
		if (idx == sis.num) e = SPI_ERR_INTERNAL;
		if (e != SPI_ERR_SUCCESS) break;
		lpInfo->crc = 0;
		lpInfo->timestamp = 0;
	} while (0);
	spi_istream_close(&sis);
	return e;
}

spi_errorcode_t PASCAL GetFile(LPSTR src, long len, LPSTR dest, unsigned int flag, FARPROC prgresisCallback, long lData)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	lpspi_ostream_t po = 0;
	spi_istream_t sis;
	spi_fileinfo_t info;
	if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED;
	do
	{
		unsigned char wavhead[0x2c];
		unsigned long frm_cur, frm_num, r;
		unsigned frm_len, idx;
		int delta, data;
		e = spi_istream_open(&sis, src, 0, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		e = getinfo(&sis, &info, len, 0);
		if (e != SPI_ERR_SUCCESS) break;
		r = info.filesize;
		frm_num = info.timestamp;
		e = spi_ostream_open(&po, dest, r, flag & 0x700);
		if (e != SPI_ERR_SUCCESS) break;
		if (!spi_istream_seek(&sis, info.crc)) { e = SPI_ERR_FILE_READ; break; }

		spi_memcpy(wavhead, wavhead_init, 0x28);
		writedword(wavhead + 0x04, r - 0x08);
		writedword(wavhead + 0x18, sis.srate);
		writedword(wavhead + 0x1c, sis.srate * 2);
		writedword(wavhead + 0x28, r - 0x2c);
		e = spi_ostream_write(po, wavhead, 0x2c);
		if (e != SPI_ERR_SUCCESS) break;

		delta = 6;
		data = 1;
		for (idx = 0; idx < 0x7f; idx++)
		{
			po->ulaw_table[0x7f - idx] = -data;
			po->ulaw_table[0x81 + idx] = +data;
			data += delta;
			delta += 4;
		}
		po->ulaw_table[0x00] = -0x8000;
		po->ulaw_table[0x80] = 0;
		po->ulaw_table[0xff] = 0x7fff;

		if (frm_num * 2 != spi_istream_read(&sis, po->frame_table, frm_num * 2)) { e = SPI_ERR_FILE_READ; break; }
		for (frm_cur = 0; frm_cur < frm_num; frm_cur++)
		{
			if (((lpspi_progresscallback_t)prgresisCallback)(frm_cur, frm_num, lData))
			{
				e = SPI_ERR_ABORT_BY_USER;
				break;
			}
			frm_len = readword(po->frame_table + frm_cur * 2);
			if (frm_len > 0x400) { e = SPI_ERR_FILE_READ; break; }
			if (frm_len != spi_istream_read(&sis, po->i_buf, frm_len)) { e = SPI_ERR_FILE_READ; break; }
			spi_memset(po->o_buf, 0, 0x400 * 2);
			if (frm_len == 0x400)
			{
				for (idx = 0; idx < 0x400; idx++)
				{
					writeword(po->o_buf + idx * 2, po->ulaw_table[po->i_buf[idx]]);
				}
			}
			else if (frm_len)
			{
				unsigned char *ip, hpos;
				unsigned char cur;
				ip = po->i_buf;
				hpos = 0;
				cur = 0;
				for (idx = 0; idx < 0x400; idx++)
				{
					delta = (*ip >> hpos) & 0xf;
					hpos ^= 4;
					if (hpos == 0) ip++;
					if (delta == 15)
					{
						delta = (*ip >> hpos) & 0xf;
						hpos ^= 4;
						if (hpos == 0) ip++;
						delta += ((*ip >> hpos) & 0xf) << 4;
						hpos ^= 4;
						if (hpos == 0) ip++;
					}
					if (delta & 1)
						cur += (delta >> 1) + 1;
					else
						cur -= (delta >> 1) + 0;
					writeword(po->o_buf + idx * 2, po->ulaw_table[cur]);
				}
			}
			e = spi_ostream_write(po, po->o_buf, 0x400 * 2);
			if (e != SPI_ERR_SUCCESS) break;
		}

		if (e != SPI_ERR_SUCCESS) break;
		if (prgresisCallback)
		{
			if (((lpspi_progresscallback_t)prgresisCallback)(frm_cur, frm_num, lData))
			{
				e = SPI_ERR_ABORT_BY_USER;
				break;
			}
		}
		spi_ostream_flush(po);
	} while (0);
	spi_ostream_close(po);
	spi_istream_close(&sis);
	return e;
}

#if defined(_MSC_VER) && !defined(_DEBUG)
#pragma comment(linker, "/STUB:../../stub.exe")
#pragma comment(linker, "/OPT:NOWIN98")
#pragma comment(linker, "/MERGE:.rdata=.text")
#pragma comment(linker, "/ENTRY:DllMain")
#endif
BOOL WINAPI DllMain(HINSTANCE hInstDLL, DWORD dwReason, LPVOID lpReserved)
{
	if (dwReason == DLL_PROCESS_ATTACH)
	{
		DisableThreadLibraryCalls(hInstDLL);
	}
	return 1;
}
