#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <time.h>
#include "../sversion.h"
#include "../libbz2/bzlib.h"


typedef unsigned char *lpspi_uchar_t;
typedef const unsigned char *lpspi_cuchar_t;
typedef int (PASCAL *lpspi_progresscallback_t)(int nNum, int nDenom, long lData);

#include <pshpack1.h>
typedef struct
{
	unsigned char method[8];    /* k@̎ */
	unsigned long position;     /* t@Cł̈ʒu */
	unsigned long compsize;     /* kꂽTCY */
	unsigned long filesize;     /* ̃t@CTCY */
	time_t timestamp;           /* t@C̍XV */
	char path[200];             /* ΃pX */
	char filename[200];         /* t@Cl[ */
	unsigned long crc;          /* CRC */
} spi_fileinfo_t, *lpspi_fileinfo_t;
#include <poppack.h>

typedef struct
{
	unsigned long ctype;
	unsigned long ofs;
} spi_fileinfoex_t, *lpspi_fileinfoex_t;

typedef struct
{
	unsigned int f;
	HANDLE file_h;
	HLOCAL *mem_r;
	HLOCAL mem_h;
	lpspi_uchar_t mem_p;
	unsigned long mem_len;
	unsigned long mem_pos;
	unsigned char bitbuf[0x4000];
	unsigned char outbuf[0x4000];
	unsigned char lzssdic[0x100];
	unsigned bitbufpos;
	unsigned bitbuflen;
	unsigned bitpos;
	unsigned bitcur;
	unsigned lzssdicpos;
} spi_ostream_t, *lpspi_ostream_t;

typedef struct
{
	unsigned int f;
	HANDLE file_h;
	/* unsigned long file_ofs; */
	lpspi_uchar_t mem_p;
	unsigned long mem_len;
	unsigned long mem_pos;
	/* --- */
	unsigned long istop;
	unsigned long ispos;
	unsigned long islen;
	unsigned char isbuf[8192];
	/* --- */
	unsigned long num;
	unsigned long headofs;
	unsigned long baseofs;
	int type;
	int hasnbzspb;
} spi_istream_t, *lpspi_istream_t;

enum
{
	SPI_ERR_SUCCESS = 0,
	SPI_ERR_NOTIMPLEMENTED = -1,
	SPI_ERR_ABORT_BY_USER = 1,
	SPI_ERR_UNKNOWN_FORMAT = 2,
	SPI_ERR_DATA_BROKEN = 3,
	SPI_ERR_SHORT_OF_MEMORY = 4,
	SPI_ERR_MEMORY_ERROR = 5,
	SPI_ERR_FILE_READ = 6,
	SPI_ERR_RESERVED = 7,
	SPI_ERR_INTERNAL = 8
};
typedef int spi_errorcode_t;

static unsigned spi_strlen(LPCSTR p)
{
	unsigned l = 0;
	while (*p) p++, l++;
	return l;
}
static unsigned spi_splitpath(LPSTR p)
{
	unsigned r,l;
	for (r = l = 0; p[l]; l++)
	{
		if (IsDBCSLeadByte(p[l])) { l++; continue; }
		if (p[l] == ':' || p[l] == '/' || p[l] == '\\')
			r = l + 1;
	}
	return r;
}
static unsigned spi_splitext(LPSTR p)
{
	unsigned r,l;
	for (r = l = 0; p[l]; l++)
	{
		if (IsDBCSLeadByte(p[l])) { l++; continue; }
		if (p[l] == '.')
			r = l;
	}
	return r;
}
static void spi_memcpy(LPVOID d, LPCVOID s, unsigned l)
{
	unsigned p;
	for (p = 0; p < l; p++) ((LPSTR)d)[p] = ((LPCSTR)s)[p];
}

static void spi_memset(LPVOID d, char c, unsigned l)
{
	unsigned p;
	for (p = 0; p < l; p++) ((LPSTR)d)[p] = c;
}


/*
static unsigned long readdword(lpspi_cuchar_t p)
{
	unsigned long ret;
	ret  = ((unsigned long)p[3]) << (3 * 8);
	ret |= ((unsigned long)p[2]) << (2 * 8);
	ret |= ((unsigned)p[1]) << (1 * 8);
	ret |= ((unsigned)p[0]) << (0 * 8);
	return ret;
}
*/

static unsigned char *writedwordle(unsigned char *p, unsigned long v)
{
	*p++ = (unsigned char)(v >> (8 * 0));
	*p++ = (unsigned char)(v >> (8 * 1));
	*p++ = (unsigned char)(v >> (8 * 2));
	*p++ = (unsigned char)(v >> (8 * 3));
	return p;
}


static int spi_istream_seek(lpspi_istream_t psis, unsigned long pos)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		if (psis->istop <= pos && pos <= psis->istop + psis->islen)
		{
			psis->ispos = pos - psis->istop;
			return 1;
		}
		else
		{
			LONG hi = 0;
			psis->ispos = psis->islen = 0;
			psis->istop = SetFilePointer(psis->file_h, pos, &hi, FILE_BEGIN);
			if (psis->istop != pos) return 0;
		}
		return 1;
	case 1:/* memory image */
		if (psis->mem_len <= pos) return 0;
		psis->mem_pos = pos;
		return 1;
	default:
		return 0;
	}
}
static unsigned int spi_istream_read(lpspi_istream_t psis, LPVOID pbuf, unsigned int len)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		{
			unsigned ret = 0;
			while (ret < len)
			{
				if (psis->islen == psis->ispos)
				{
					DWORD rl = 0;
					psis->istop += psis->islen;
					psis->ispos = psis->islen = 0;
					if (!ReadFile(psis->file_h, psis->isbuf, sizeof(psis->isbuf), &rl, NULL)) return 0;
					psis->islen = rl;
					if (rl == 0) break;
				}
				if (psis->islen > psis->ispos)
				{
					unsigned cpl = psis->islen - psis->ispos;
					if (cpl > len - ret) cpl = len - ret;
					spi_memcpy((lpspi_uchar_t)pbuf + ret, psis->isbuf + psis->ispos, cpl);
					psis->ispos += cpl;
					ret += cpl;
				}
			}
			return ret;
		}
	case 1:/* memory image */
		if (len > psis->mem_len - psis->mem_pos) len = psis->mem_len - psis->mem_pos;
		if (len) spi_memcpy(pbuf, psis->mem_p + psis->mem_pos, len);
		psis->mem_pos += len;
		return len;
	default:
		return 0;
	}
}
static void spi_istream_close(lpspi_istream_t psis)
{
	switch (psis->f)
	{
	case 0:/* disk file */
		if (psis->file_h != INVALID_HANDLE_VALUE) CloseHandle(psis->file_h);
		break;
	case 1:/* memory image */
		break;
	default:
		break;
	}
}

enum
{
	TYPE_MIN,
	TYPE_SAR = TYPE_MIN,
	TYPE_NSA,
	TYPE_NS2,
	TYPE_MAX
};

enum
{
	CTYPE_RAW = 0,
	CTYPE_SPB = 1,
	CTYPE_LZSS = 2,
	CTYPE_NBZ = 4,
	CTYPE_MAX = 4,
	CTYPE_MIZUIROWAVPATCH = 5
};

static unsigned readbytebe(unsigned char *v, lpspi_cuchar_t p, unsigned i, unsigned l)
{
	unsigned char ret = 0;
	if (i < l) ret  = p[i++];
	if (v) *v = ret;
	return i;
}

static unsigned readwordbe(unsigned *v, lpspi_cuchar_t p, unsigned i, unsigned l)
{
	unsigned ret = 0;
	if (i < l) ret  = p[i++] << 8;
	if (i < l) ret |= p[i++];
	if (v) *v = ret;
	return i;
}
static unsigned readdwordbe(unsigned long *v, lpspi_cuchar_t p, unsigned i, unsigned l)
{
	unsigned long ret = 0;
	if (i < l) ret  = ((unsigned long)p[i++]) << 24;
	if (i < l) ret |= ((unsigned long)p[i++]) << 16;
	if (i < l) ret |= ((unsigned)p[i++]) << 8;
	if (i < l) ret |= p[i++];
	if (v) *v = ret;
	return i;
}

typedef struct
{
	unsigned long ofs;
	unsigned long compsize;
	unsigned long filesize;
	unsigned long ctype;
	char filename[200];         /* t@Cl[ */
} nsafileinfo_t;
static unsigned readfileinfo(nsafileinfo_t *v, lpspi_cuchar_t p, unsigned i, unsigned l, int type)
{
	unsigned len = 0;
	char *name = 0;
	unsigned char uc, ctype;
	unsigned long ofs, filesize, compsize;
	if (v)
	{
		v->ofs = 0;
		v->filesize = 0;
		v->compsize = 0;
		v->ctype = 0;
		name = v->filename;
	}
	while (i < l)
	{
		i = readbytebe(&uc, p, i, l);
		if (uc == 0) break;
		if (name) name[len++] = (char)uc;
	}
	if (name) name[len] = '\0';
	if (type == TYPE_NS2 || type == TYPE_NSA)
		i = readbytebe(&ctype, p, i, l);
	else
		ctype = 0;
	i = readdwordbe(&ofs, p, i, l);
	i = readdwordbe(&compsize, p, i, l);
	if (type == TYPE_NS2 || type == TYPE_NSA)
		i = readdwordbe(&filesize, p, i, l);
	else
		filesize = compsize;
	if (filesize == 0) ctype = CTYPE_RAW;
	if (v)
	{
		v->ofs = ofs;
		v->filesize = filesize;
		v->compsize = compsize;
		v->ctype = ctype;
	}
	return i;
}

static int typecheck(lpspi_cuchar_t p, unsigned l, lpspi_istream_t psis)
{
	int type;
	for (type = 0; type < TYPE_MAX; type++)
	{
		unsigned i = 0;
		unsigned num;
		unsigned long baseofs;
		nsafileinfo_t nsa_fileinfo[2];
		if (type == TYPE_NS2) i = readbytebe(0, p, i, l);
		i = readwordbe(&num, p, i, l);
		if (num < 2) continue;
		i = readdwordbe(&baseofs, p, i, l);
		if (baseofs == 0) continue;
		if (baseofs >= 0x01000000) continue;
		if (type == TYPE_NS2) baseofs++;
		i = readfileinfo(nsa_fileinfo + 0, p, i, l, type);
		i = readfileinfo(nsa_fileinfo + 1, p, i, l, type);
		/* mF */
		if (nsa_fileinfo[0].ofs != 0) continue;
		if (nsa_fileinfo[0].compsize == 0) continue;
		if (nsa_fileinfo[0].ctype > CTYPE_MAX) continue;
		if (nsa_fileinfo[1].compsize == 0) continue;
		if (nsa_fileinfo[1].ctype > CTYPE_MAX) continue;
		if (nsa_fileinfo[0].compsize != nsa_fileinfo[1].ofs) continue;
		if (num > 2)
		{
			/* ߉ */
			unsigned long s0 = nsa_fileinfo[0].compsize;
			i = readfileinfo(nsa_fileinfo + 0, p, i, l, type);
			if (nsa_fileinfo[0].compsize == 0) continue;
			if (nsa_fileinfo[0].ctype > CTYPE_MAX) continue;
			if (s0 + nsa_fileinfo[1].compsize != nsa_fileinfo[0].ofs) continue;
		}
		if (psis)
		{
			psis->num = num;
			psis->headofs = (type == TYPE_NS2) ? 7 : 6;
			psis->baseofs = baseofs;
			psis->type = type;
		}
		return 1;
	}
	return 0;
}

static spi_errorcode_t spi_istream_open(lpspi_istream_t psis, LPSTR buf, long len, unsigned int flag)
{
	psis->f = flag;
	switch (psis->f)
	{
	case 0:/* disk file */
		psis->file_h = CreateFile(buf, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (psis->file_h == INVALID_HANDLE_VALUE) return SPI_ERR_FILE_READ;
		if (len != 0) return SPI_ERR_NOTIMPLEMENTED;
		/* psis->file_ofs = len */;
		/* --- */
		psis->ispos = 0;
		psis->islen = 0;
		psis->istop = 0;
		break;
	case 1:/* memory image */
		psis->mem_p = buf;
		psis->mem_len = len;
		psis->mem_pos = 0;
		break;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
	{
		unsigned char head[512];
		psis->hasnbzspb = 0;
		spi_memset(head, 0, sizeof(head));
#if 1
		spi_istream_read(psis, head, 512);
#else
		if (512 != spi_istream_read(psis, head, 512)) return SPI_ERR_FILE_READ;
#endif
		if (!typecheck(head, 512, psis)) return SPI_ERR_UNKNOWN_FORMAT;
	}
	return SPI_ERR_SUCCESS;
}

static const char method_raw[8] = "raw\0\0\0\0\0";
static const char method_spb[8] = "spb\0\0\0\0\0";
static const char method_nbz[8] = "nbz\0\0\0\0\0";
static const char method_lzss[8] = "lzss\0\0\0\0";
static const char method_unknown[8] = "unknown\0";
static const char ext_nbz[] = ".nbz";
static const char ext_wav[] = ".wav";
static const char ext_spb[] = ".spb";
static const char ext_bmp[] = ".bmp";
static const char mizuirowavpatchname[] = "w\\w4\\3\\mutuki_1367.wav";

static spi_errorcode_t getinfo(lpspi_istream_t psis, lpspi_fileinfo_t pinfo, lpspi_fileinfoex_t pinfoex, unsigned long pos, unsigned *pinfolen)
{
	unsigned long hl, hi,p,l;
	unsigned char hbuf[256];
	nsafileinfo_t nsa_fi;
	if (!spi_istream_seek(psis, pos)) return SPI_ERR_FILE_READ;
	hl = spi_istream_read(psis, hbuf, sizeof(hbuf));
	if (hl < 8) return SPI_ERR_FILE_READ;
	hi = readfileinfo(&nsa_fi, hbuf, 0, hl, psis->type);
	if (pinfolen) *pinfolen = hi;

	switch (nsa_fi.ctype)
	{
	case CTYPE_RAW:	/* raw */
		spi_memcpy(pinfo->method, method_raw, 8);
		break;
	case CTYPE_LZSS:	/* lzss */
		spi_memcpy(pinfo->method, method_lzss, 8);
		break;
	case CTYPE_NBZ:	/* nbz */
		spi_memcpy(pinfo->method, method_nbz, 8);
		nsa_fi.ctype = CTYPE_NBZ;
		break;
	case CTYPE_SPB:	/* spb */
		spi_memcpy(pinfo->method, method_spb, 8);
		break;
	default:	/* unknown */
		spi_memcpy(pinfo->method, method_unknown, 8);
		/* fʂ */
		nsa_fi.ctype = CTYPE_RAW;
		nsa_fi.filesize = nsa_fi.compsize;
		break;
	}

	if (lstrcmp(nsa_fi.filename, mizuirowavpatchname) == 0)
	{
		nsa_fi.ctype = CTYPE_MIZUIROWAVPATCH;
		nsa_fi.filesize += 12;
	}

	p = spi_splitpath(nsa_fi.filename);
	l = spi_strlen(nsa_fi.filename) + 1;
	spi_memcpy(pinfo->path, nsa_fi.filename, p);
	pinfo->path[p] = 0;
	spi_memcpy(pinfo->filename, nsa_fi.filename + p, l - p);
	l = spi_splitext(pinfo->filename);
	if (l && lstrcmpi(pinfo->filename + l, ext_nbz) == 0)
	{
		psis->hasnbzspb++;
		pinfo->timestamp = 0;
		pinfo->crc = psis->baseofs + nsa_fi.ofs;
	}
	else if (l && lstrcmpi(pinfo->filename + l, ext_spb) == 0)
	{
		psis->hasnbzspb++;
		pinfo->timestamp = 1;
		pinfo->crc = psis->baseofs + nsa_fi.ofs;
	}
	else
	{
		pinfo->timestamp = 0;
		pinfo->crc = 0;
	}


	pinfo->compsize = nsa_fi.compsize;
	pinfo->filesize = nsa_fi.filesize;
	if (pinfoex)
	{
		pinfoex->ofs = psis->baseofs + nsa_fi.ofs;
		pinfoex->ctype = nsa_fi.ctype;
	}
	pinfo->position = pos;
	return SPI_ERR_SUCCESS;
}
static spi_errorcode_t getinfoext(lpspi_istream_t psis, lpspi_fileinfo_t pinfo)
{
	unsigned w, l;
	unsigned long ul;
	unsigned char buf[4];
	if (pinfo->crc == 0) return SPI_ERR_SUCCESS;
	if (!spi_istream_seek(psis, pinfo->crc)) return SPI_ERR_FILE_READ;
	if (4 != spi_istream_read(psis, buf, 4)) return SPI_ERR_FILE_READ;
	if (pinfo->timestamp)
	{
		readwordbe(&w, buf, 0, 4);
		readwordbe(&l, buf, 2, 4);
		w = w * 3 + ((4 - ((w * 3) & 3)) & 3);
		ul = 0x36 + w * l;
		lstrcat(pinfo->filename, ext_bmp);
	}
	else
	{
		readdwordbe(&ul, buf, 0, 4);
		lstrcat(pinfo->filename, ext_wav);
	}
	pinfo->filesize = ul;
	return SPI_ERR_SUCCESS;
}

static spi_errorcode_t spi_ostream_open(lpspi_ostream_t *ppo, LPSTR buf, unsigned long len, unsigned int flag)
{
	lpspi_ostream_t po;
	*ppo = po = LocalAlloc(LPTR, sizeof(spi_ostream_t));
	if (po == NULL) return SPI_ERR_SHORT_OF_MEMORY;
	po->lzssdicpos = 0;
	po->f = flag >> 8;
	switch (po->f)
	{
	case 0:/* disk file */
		po->file_h = CreateFile(buf, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (po->file_h == INVALID_HANDLE_VALUE) return SPI_ERR_FILE_READ;
		break;
	case 1:/* memory image */
		po->mem_r = (HLOCAL *)buf;
		*po->mem_r = 0;
		po->mem_p = 0;
		po->mem_h = LocalAlloc(LMEM_MOVEABLE, len ? len : 1);
		if (po->mem_h == NULL) return SPI_ERR_SHORT_OF_MEMORY;
		po->mem_p = (lpspi_uchar_t)LocalLock(po->mem_h);
		if (po->mem_p == NULL) return SPI_ERR_MEMORY_ERROR;
		po->mem_len = len;
		po->mem_pos = 0;
		break;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
	return SPI_ERR_SUCCESS;
}
static spi_errorcode_t spi_ostream_write(lpspi_ostream_t po, LPSTR buf, unsigned long len)
{
	switch (po->f)
	{
	case 0:/* disk file */
		{
			DWORD r = 0;
			if (!WriteFile(po->file_h, buf, len, &r, NULL)) return SPI_ERR_FILE_READ;
		}
		return SPI_ERR_SUCCESS;
	case 1:/* memory image */
		if (len > po->mem_len - po->mem_pos) len = po->mem_len - po->mem_pos;
		if (len) spi_memcpy(po->mem_p + po->mem_pos, buf, len);
		po->mem_pos += len;
		return SPI_ERR_SUCCESS;
	default:
		return SPI_ERR_NOTIMPLEMENTED;
	}
}

static void spi_ostream_lzss_init(lpspi_ostream_t po)
{
	int i;
	for (i = 0; i < sizeof(po->lzssdic); i++) po->lzssdic[i] = 0x20;
	po->bitbufpos = 0;
	po->bitbuflen = 0;
	po->bitpos = 0;
}
static unsigned spi_ostream_lzss_getc(lpspi_ostream_t po, lpspi_istream_t psis)
{
	if (po->bitbufpos >= po->bitbuflen)
	{
		po->bitbuflen = spi_istream_read(psis, po->bitbuf, sizeof(po->bitbuf));
		po->bitbufpos = 0;
	}
	return po->bitbuf[po->bitbufpos++];
}
static unsigned spi_ostream_lzss_getbit(lpspi_ostream_t po, lpspi_istream_t psis)
{
	if (po->bitpos == 0)
	{
		po->bitcur = spi_ostream_lzss_getc(po, psis);
		po->bitpos = 8;
	}
	return (po->bitcur >> --po->bitpos) & 1;
}
static unsigned spi_ostream_lzss_getbits(lpspi_ostream_t po, lpspi_istream_t psis, unsigned len)
{
	unsigned ret = 0;
	do
	{
		ret += ret;
		ret += spi_ostream_lzss_getbit(po, psis);
	} while (--len);
	return ret;
}
static void spi_ostream_lzss_putc(lpspi_ostream_t po, unsigned c)
{
	po->lzssdic[po->lzssdicpos] = c;
	if (++po->lzssdicpos == 0x100)
	{
		spi_ostream_write(po, po->lzssdic, 0x100);
		po->lzssdicpos = 0;
	}
}
static void spi_ostream_lzss_copy(lpspi_ostream_t po, unsigned cpos, unsigned clen)
{
	while (clen--)
	{
		spi_ostream_lzss_putc(po, po->lzssdic[cpos]);
		cpos = (cpos + 1) & 0xff;
	}
}
static void spi_ostream_flush(lpspi_ostream_t po)
{
	if (po->lzssdicpos) spi_ostream_write(po, po->lzssdic, po->lzssdicpos);
	switch (po->f)
	{
	case 0:/* disk file */
		break;
	case 1:/* memory image */
		LocalUnlock(po->mem_h);
		*po->mem_r = po->mem_h;
		po->mem_p = NULL;
		po->mem_h = NULL;
		break;
	}
}
static void spi_ostream_close(lpspi_ostream_t po)
{
	if (po)
	{
		switch (po->f)
		{
			case 0:/* disk file */
				if (po->file_h != INVALID_HANDLE_VALUE) CloseHandle(po->file_h);
				break;
			case 1:/* memory image */
				if (po->mem_p != NULL) LocalUnlock(po->mem_h);
				if (po->mem_h != NULL) LocalFree(po->mem_h);
				break;
		}
		LocalFree(po);
	}
}
SUSIEPLUGIN_INFO_NS2(spi_info)
int PASCAL GetPluginInfo(int infono, LPSTR buf, int buflen)
{
	int l;
	if (((unsigned)infono) >= (sizeof(spi_info) / sizeof(const char *))) return 0;
	l = spi_strlen(spi_info[infono]) + 1;
	if (l > buflen) l = buflen;
	spi_memcpy(buf, spi_info[infono], l);
	return l;
}

int PASCAL IsSupported(LPSTR filename,DWORD dw)
{
	int ret = 0;
	LONG hi = 0;
	HLOCAL pbuf = NULL;
	do
	{
		lpspi_uchar_t pdata;
		if (HIWORD(((DWORD)dw)))
		{
			pdata = (lpspi_uchar_t)dw;
		}
		else
		{
			HFILE h;
			pbuf = LocalAlloc(LPTR, 2048);
			if (pbuf == NULL) break;
			h = (HFILE)LOWORD(((DWORD)dw));
			if (0 != SetFilePointer((HANDLE)h, 0, &hi, FILE_BEGIN)) break;
			ReadFile((HANDLE)h, pbuf, 2048, &hi, NULL);
			pdata = (lpspi_uchar_t)pbuf;
		}
		ret = typecheck(pdata, 2048, 0) != 0;
	} while (0);
	if (pbuf != NULL) LocalFree(pbuf);
	return ret;
}

spi_errorcode_t PASCAL GetArchiveInfo(LPSTR buf, long len, unsigned int flag, HLOCAL *lphInf)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	spi_istream_t sis;
	HLOCAL hinfo = NULL;
	lpspi_fileinfo_t pinfo = NULL;
	/* if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED; */
	do
	{
		unsigned pos, infolen, num;
		e = spi_istream_open(&sis, buf, len, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		pos = sis.headofs;
		num = sis.num;
		hinfo = LocalAlloc(LHND, sis.num * sizeof(spi_fileinfo_t) + 1);
		if (hinfo == NULL) { e = SPI_ERR_SHORT_OF_MEMORY; break;}
		pinfo = (lpspi_fileinfo_t)LocalLock(hinfo);
		if (pinfo == NULL) { e = SPI_ERR_MEMORY_ERROR; break;}
		while (num)
		{
			e = getinfo(&sis, pinfo, 0, pos, &infolen);
			if (e != SPI_ERR_SUCCESS) break;
			pinfo++;
			num--;
			pos += infolen;
		}
		if (e != SPI_ERR_SUCCESS) break;
		pinfo->method[0] = 0;
		if (sis.hasnbzspb)
		{
			num = sis.num;
			pinfo -= sis.num;
			while (num)
			{
				e = getinfoext(&sis, pinfo);
				if (e != SPI_ERR_SUCCESS) break;
				pinfo->crc = 0;
				pinfo->timestamp = 0;
				pinfo++;
				num--;
			}
		}
		LocalUnlock(hinfo);
		*lphInf = hinfo;
		pinfo = NULL; hinfo = NULL;
	} while (0);
	if (pinfo != NULL) LocalUnlock(hinfo);
	if (hinfo != NULL) LocalFree(hinfo);
	spi_istream_close(&sis);
	return e;
}

spi_errorcode_t PASCAL GetFileInfo(LPSTR buf, long len, LPSTR filename, unsigned int flag, lpspi_fileinfo_t lpInfo)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	spi_istream_t sis;
	/* if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED; */
	do
	{
		unsigned pos, infolen, num;
		unsigned char fn[200+200];
		e = spi_istream_open(&sis, buf, len, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		pos = sis.headofs;
		num = sis.num;
		while (num)
		{
			e = getinfo(&sis, lpInfo, 0, pos, &infolen);
			if (e != SPI_ERR_SUCCESS) break;
			lstrcpy(fn, lpInfo->path);
			lstrcat(fn, lpInfo->filename);
			if (lpInfo->crc) lstrcat(fn, lpInfo->timestamp ? ext_bmp : ext_wav);
			if ((((flag & 0x80)?lstrcmpi:lstrcmp)(fn, filename)) == 0) break;
			num--;
			pos += infolen;
		}
		if (num == 0) e = SPI_ERR_INTERNAL;
		if (e != SPI_ERR_SUCCESS) break;
		e = getinfoext(&sis, lpInfo);
		if (e != SPI_ERR_SUCCESS) break;
		lpInfo->crc = 0;
		lpInfo->timestamp = 0;
	} while (0);
	spi_istream_close(&sis);
	return e;
}

static void * bzalloc(void *opaque, int items, int size)
{
	return LocalAlloc(LPTR, items * size);
}
static void bzfree(void* opaque, void* addr)
{
	LocalFree(addr);
}


spi_errorcode_t PASCAL GetFile(LPSTR src, long len, LPSTR dest, unsigned int flag, FARPROC prgresisCallback, long lData)
{
	spi_errorcode_t e = SPI_ERR_SUCCESS;
	lpspi_ostream_t po = 0;
	unsigned char *spbbuf = 0;
	spi_istream_t sis;
	spi_fileinfo_t info;
	spi_fileinfoex_t infoex;
	bz_stream bzs;
	unsigned long p, r, l, pcur, pmax;
	if ((flag & 7) != 0) return SPI_ERR_NOTIMPLEMENTED;
	do
	{
		e = spi_istream_open(&sis, src, 0, flag & 7);
		if (e != SPI_ERR_SUCCESS) break;
		e = getinfo(&sis, &info, &infoex, len, 0);
		if (e != SPI_ERR_SUCCESS) break;
		e = getinfoext(&sis, &info);
		if (e != SPI_ERR_SUCCESS) break;
		r = info.filesize;
		e = spi_ostream_open(&po, dest, r, flag & 0x700);
		if (e != SPI_ERR_SUCCESS) break;
		if (info.crc) infoex.ctype = CTYPE_NBZ;
		if (info.timestamp) infoex.ctype = CTYPE_SPB;
		if (!spi_istream_seek(&sis, infoex.ofs + ((infoex.ctype == CTYPE_NBZ) ? 4 : 0))) { e = SPI_ERR_FILE_READ; break; }
		p = 0;
		pcur = ~0;
		pmax = (r + ((1 >> 16) - 1)) >> 16;

		if (infoex.ctype == CTYPE_MIZUIROWAVPATCH)
		{
			unsigned char buf[12 + 8];
			l = spi_istream_read(&sis, buf + 12, 8);
			if (l < 8)
			{
				e = SPI_ERR_FILE_READ;
				break;
			}
			writedwordle(buf + 0, 0x46464952);
			writedwordle(buf + 4, r - 8);
			writedwordle(buf + 8, 0x45564157);
			writedwordle(buf + 16, 0x10);
			e = spi_ostream_write(po, buf, 12 + 8);
			if (e != SPI_ERR_SUCCESS) break;
			infoex.ctype = CTYPE_RAW;
			p = 12 + 8;
		}
		switch (infoex.ctype)
		{
		case CTYPE_SPB:
			{
				unsigned char *desp;
				unsigned width, height, bwidth;
				spi_ostream_lzss_init(po);
				width  = spi_ostream_lzss_getbits(po, &sis, 16);
				height = spi_ostream_lzss_getbits(po, &sis, 16);
				bwidth = width * 3 + ((4 - ((width * 3) & 3)) & 3);
				spbbuf = bzalloc(0, r, 1);
				pmax = 3 * height;
				if (spbbuf == 0)
				{
					e = SPI_ERR_SHORT_OF_MEMORY;
					break;
				}
				spbbuf[0] = 0x42;
				spbbuf[1] = 0x4d;
				desp = writedwordle(spbbuf + 2, bwidth * height + 0x36);
				desp = writedwordle(desp, 0);
				desp = writedwordle(desp, 0x36);
				desp = writedwordle(desp, 0x28);
				desp = writedwordle(desp, width);
				desp = writedwordle(desp, height);
				desp = writedwordle(desp, 0x00180001);
				desp = writedwordle(desp, 0);
				desp = writedwordle(desp, bwidth * height);
				desp = writedwordle(desp, 0);
				desp = writedwordle(desp, 0);
				desp = writedwordle(desp, 0);
				desp = writedwordle(desp, 0);
				for (p = 0; p < 3; p++)
				{
					unsigned x, y;
					unsigned char curcol, copylen, diflen, difmode;
					curcol = spi_ostream_lzss_getbits(po, &sis, 8);
					copylen = 1;
					diflen = 0;
					for (y = 0; y < height; y++)
					{
						if (prgresisCallback)
						{
							if (((lpspi_progresscallback_t)prgresisCallback)(p *  height + y, pmax, lData))
							{
								e = SPI_ERR_ABORT_BY_USER;
								break;
							}
						}
						desp = spbbuf + 0x36 + p + (bwidth * (height - y - 1));
						x = 0;
						while (x < width)
						{
							if (copylen)
							{
								copylen--;
							}
							else if (diflen)
							{
								diflen--;
								if (difmode == 8)
								{
									curcol = spi_ostream_lzss_getbits(po, &sis, 8);
								}
								else
								{
									unsigned dif;
									dif = spi_ostream_lzss_getbits(po, &sis, difmode);
									if (dif & 1)
										curcol += (dif >> 1) + 1;
									else
										curcol -= (dif >> 1);
								}
							}
							else
							{
								difmode = spi_ostream_lzss_getbits(po, &sis, 3);
								if (difmode != 0)
								{
									if (difmode == 7)
										difmode = spi_ostream_lzss_getbits(po, &sis, 1) + 1;
									else
										difmode += 2;
									diflen = 4;
									continue;
								}
								copylen = 3;
							}
							if (y & 1)
								desp[(width - x - 1) * 3] = curcol;
							else
								desp[x * 3] = curcol;
							x++;
						}
					}
					while (diflen)
					{
						diflen--;
						spi_ostream_lzss_getbits(po, &sis, difmode);
					}
				}
				if (e != SPI_ERR_SUCCESS) break;
				e = spi_ostream_write(po, spbbuf, r);
				if (e != SPI_ERR_SUCCESS) break;
			}
			break;
		case CTYPE_NBZ:
			bzs.bzalloc = bzalloc;
			bzs.bzfree = bzfree;
			bzs.opaque = 0;
			if (BZ2_bzDecompressInit(&bzs, 0, 0) != BZ_OK)
			{
				e = SPI_ERR_FILE_READ;
				break;
			}
			bzs.next_in = po->bitbuf;
			bzs.avail_in = spi_istream_read(&sis, po->bitbuf, sizeof(po->bitbuf));
			bzs.next_out = po->outbuf;
			bzs.avail_out = sizeof(po->outbuf);
			while (p < r)
			{
				int ret;
				if (prgresisCallback && pcur != (p >> 16))
				{
					pcur = (p >> 16);
					if (((lpspi_progresscallback_t)prgresisCallback)(pcur, pmax, lData))
					{
						e = SPI_ERR_ABORT_BY_USER;
						break;
					}
				}
				ret = BZ2_bzDecompress(&bzs);
				if (ret < 0)
				{
					e = SPI_ERR_FILE_READ;
					break;
				}
				l = sizeof(po->outbuf) - bzs.avail_out;
				if (l)
				{
					e = spi_ostream_write(po, po->outbuf, l);
					if (e != SPI_ERR_SUCCESS) break;
					bzs.next_out = po->outbuf;
					bzs.avail_out = sizeof(po->outbuf);
					p += l;
				}
				if (ret == BZ_STREAM_END) break;
				if (0 == bzs.avail_in)
				{
					bzs.next_in = po->bitbuf;
					bzs.avail_in = spi_istream_read(&sis, po->bitbuf, sizeof(po->bitbuf));
				}
			}
			BZ2_bzDecompressEnd(&bzs);
			break;
		case CTYPE_RAW:
			while (p < r)
			{
				if (prgresisCallback && pcur != (p >> 16))
				{
					pcur = (p >> 16);
					if (((lpspi_progresscallback_t)prgresisCallback)(pcur, pmax, lData))
					{
						e = SPI_ERR_ABORT_BY_USER;
						break;
					}
				}
				l = spi_istream_read(&sis, po->bitbuf, ((r - p) > sizeof(po->bitbuf)) ? sizeof(po->bitbuf) : (r - p));
				if (l == 0) { e = SPI_ERR_FILE_READ; break; }
				e = spi_ostream_write(po, po->bitbuf, l);
				if (e != SPI_ERR_SUCCESS) break;
				p += l;
			}
			break;
		case CTYPE_LZSS:
			{
				unsigned c, cpos, clen;
				spi_ostream_lzss_init(po);
				while (p < r)
				{
					if (prgresisCallback && pcur != (p >> 16))
					{
						pcur = (p >> 16);
						if (((lpspi_progresscallback_t)prgresisCallback)(pcur, pmax, lData))
						{
							e = SPI_ERR_ABORT_BY_USER;
							break;
						}
					}
					if (spi_ostream_lzss_getbits(po, &sis, 1))
					{
						c = spi_ostream_lzss_getbits(po, &sis, 8);
						spi_ostream_lzss_putc(po, c);
						p++;
					}
					else
					{
						cpos = (spi_ostream_lzss_getbits(po, &sis, 8) + 17) & 255;
						clen  = spi_ostream_lzss_getbits(po, &sis, 4) + 2;
						if (clen > r - p) clen = r - p;
						spi_ostream_lzss_copy(po, cpos, clen);
						p += clen;
					}
				}
			}
			break;
		default:
			e = SPI_ERR_UNKNOWN_FORMAT;
			break;
		}
		if (e != SPI_ERR_SUCCESS) break;
		if (prgresisCallback)
		{
			if (((lpspi_progresscallback_t)prgresisCallback)(pmax, pmax, lData))
			{
				e = SPI_ERR_ABORT_BY_USER;
				break;
			}
		}
		spi_ostream_flush(po);
	} while (0);
	if (spbbuf) bzfree(0, spbbuf);
	spi_ostream_close(po);
	spi_istream_close(&sis);
	return e;
}

#if defined(_MSC_VER) && !defined(_DEBUG)
#pragma comment(linker, "/STUB:../../stub.exe")
#pragma comment(linker, "/OPT:NOWIN98")
#pragma comment(linker, "/MERGE:.rdata=.text")
#pragma comment(linker, "/ENTRY:DllMain")
#endif
BOOL WINAPI DllMain(HINSTANCE hInstDLL, DWORD dwReason, LPVOID lpReserved)
{
	if (dwReason == DLL_PROCESS_ATTACH)
	{
		DisableThreadLibraryCalls(hInstDLL);
	}
	return 1;
}
