#include "foobar2000/SDK/input.h"
#include "foobar2000/SDK/componentversion.h"

#include "../libpcm/libpcm.h"
#include "../libpcm/fader.h"
#include "../pversion.h"

#include "reader_foo.h"

FOOBAR2000COMPONENT_EXTS(foobar2000_extlist);

#define UNIT_RENDER 4096

class input_okiadpcm : public input_pcm
{
private:
	mem_block_t<char> buffer;
	LIBPCM_DECODER *decoder;
	LIBPCM_FADER *fader;
public:
	input_okiadpcm()
	{
		decoder = 0;
		fader = 0;
	}
	~input_okiadpcm()
	{
		if (fader) libpcm_fader_terminate(fader);
		if (decoder) libpcm_close(decoder);
	}
	virtual bool test_filename(const char * fn,const char * ext) 
	{
		int l = lstrlen(fn);
		for (int i = 0; foobar2000_extlist[i]; i++)
			if (!stricmp_utf8(ext, foobar2000_extlist[i])) return true;
		return false;
	}
	bool open(reader * r,file_info * info,unsigned flags)
	{
		string_ansi_from_utf8 fn_ansi(info->get_file_path());
		libpcm_IReader *preader = libpcm_CreateFooReader(r, fn_ansi);
		if (!preader) return false;
		decoder = libpcm_open_from_stream(preader);
		preader->lpVtbl->Release(preader);
		if (!decoder) return false;
		fader = libpcm_fader_initialize(decoder);
		if (!fader) return false;
		if (flags & OPEN_FLAG_NO_LOOPING)
			libpcm_fader_configure(fader, 1, 0);
		else
			libpcm_fader_configure(fader, 2, 5000);
		if (flags & OPEN_FLAG_GET_INFO)
		{
			info->set_length(double(libpcm_fader_get_length_ms(fader)) / 1000);
			info->info_set_int("bitrate",  libpcm_get_bitrate(decoder) / 1000);
			info->info_set_int("samplerate", libpcm_get_samplerate(decoder));
			info->info_set_int("channels", libpcm_get_numberofchannels(decoder));
			info->info_set_int("bitspersample", libpcm_get_bitspersample(decoder));
			info->info_set("codec", libpcm_get_codecname(libpcm_get_codec(decoder)));
		}
		return true;
	}
#if (defined(UTF8API_VERSION) && (UTF8API_VERSION >= 2))
	virtual set_info_t set_info(reader *r,const file_info * info)
	{
		return SET_INFO_FAILURE;
	}
#else
	virtual int set_info(reader *r,const file_info * info)
	{
		return 0;
	}
#endif
	virtual int get_samples_pcm(void ** out_buffer,int * out_size,int * srate,int * bps,int * nch)
	{
		if (!decoder) return 0;
		uint_t ba = libpcm_get_blockalign(decoder);
		char * ptr = buffer.set_size(ba * UNIT_RENDER);
		uint_t numread = libpcm_fader_read(fader, ptr, UNIT_RENDER);
		if (!numread) return 0;
		*out_buffer = ptr;
		*out_size = ba * numread;
		*srate = libpcm_get_samplerate(decoder);
		*bps = libpcm_get_bitspersample(decoder);
		*nch = libpcm_get_numberofchannels(decoder);
		return 1;
	}
	virtual bool seek(double seconds)
	{
		if (!decoder) return false;
		libpcm_seek_ms(decoder, uint32_t(seconds * 1000));
		return true;
	}
};

static service_factory_t<input,input_okiadpcm> foo;
DECLARE_COMPONENT_VERSION(FOOBAR2000COMPONENT_NAME,FOOBAR2000COMPONENT_VERSION,0);
