#define WIN32_LEAN_AND_MEAN
#define NOGDI
#include <windows.h>
#include "kobarin/kmp_pi.h"

#include "../libpcm/libpcm.h"
#include "../pversion.h"

#define UNIT_RENDER 1024
#define kbmediaplugin_version KBMEDIAPLUGIN_VERSION
static const char kbmediaplugin_copyright[] = KBMEDIAPLUGIN_COPYRIGHT;
static const char kbmediaplugin_discription[] = KBMEDIAPLUGIN_NAME;
KBMEDIAPLUGIN_EXTS(kbmediaplugin_exts);

static HKMP WINAPI kmp_Open(const char *cszFileName, SOUNDINFO *pInfo)
{
	LIBPCM_DECODER *plibpcm;
	if (!cszFileName || !cszFileName[0] || !pInfo) return 0;
	plibpcm = libpcm_open_from_file(cszFileName);
	if(!plibpcm) return 0;
	pInfo->dwSamplesPerSec = libpcm_get_samplerate(plibpcm);
	pInfo->dwChannels = libpcm_get_numberofchannels(plibpcm);
	pInfo->dwBitsPerSample = libpcm_get_bitspersample(plibpcm);
	pInfo->dwLength = libpcm_get_looplength(plibpcm) ? (0 * libpcm_get_looplength_ms(plibpcm) + libpcm_get_length_ms(plibpcm)) : libpcm_get_length_ms(plibpcm);
	pInfo->dwSeekable = 1;
	pInfo->dwUnitRender = libpcm_get_blockalign(plibpcm) * UNIT_RENDER;
	pInfo->dwReserved1 = libpcm_get_looplength(plibpcm) ? 1 : 0;
	pInfo->dwReserved2 = 0;
	if (pInfo->dwReserved1) libpcm_switch_loop(plibpcm, 1);
	return (HKMP)plibpcm;
}

static void WINAPI kmp_Close(HKMP hKMP)
{
	LIBPCM_DECODER *plibpcm = hKMP;
	if (plibpcm) libpcm_close(plibpcm);
}

static DWORD WINAPI kmp_Render(HKMP hKMP, BYTE* Buffer, DWORD dwSize)
{
	uint_t ba;
	LIBPCM_DECODER *plibpcm = hKMP;
	if(!plibpcm || !Buffer) return 0;
	ba = libpcm_get_blockalign(plibpcm);
	return libpcm_read(plibpcm, Buffer, dwSize / ba) * ba;
}

static DWORD WINAPI kmp_SetPosition(HKMP hKMP, DWORD dwPos)
{
	LIBPCM_DECODER *plibpcm = hKMP;
	if(!plibpcm) return 0;
	libpcm_seek_ms(plibpcm, dwPos);
	return dwPos;
}

KMPMODULE* LIBPCM_CDECL KMP_GETMODULE(void)
{
	static const KMPMODULE Mod = {
		KMPMODULE_VERSION,
		kbmediaplugin_version,
		kbmediaplugin_copyright,
		kbmediaplugin_discription,
		(const char **)kbmediaplugin_exts,
		1,	/* dwReentrant */
		0,	/* Init */
		0,	/* Deinit */
		kmp_Open,
		0,	/* OpenFromBuffer */
		kmp_Close,
		kmp_Render,
		kmp_SetPosition
	};
	return (KMPMODULE *)&Mod;
}

#if defined(_MSC_VER) && !defined(_DEBUG)
#pragma comment(linker,"/OPT:NOWIN98")
#pragma comment(linker,"/ENTRY:DllMain")
#pragma comment(linker,"/MERGE:.rdata=.text")
#pragma comment(linker,"/STUB:../stub.exe")
#endif

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		DisableThreadLibraryCalls(hinstDLL);
	}
#ifdef _DEBUG
	else if (fdwReason == DLL_PROCESS_DETACH)
	{
		extern int LIBPCM_API libpcm_leakcheck_dbg(void);
		if (libpcm_leakcheck_dbg())
		{
			MessageBoxA(0,"libpcm","leak",MB_OK);
		}
	}
#endif
	return TRUE;
}
