#define WIN32_LEAN_AND_MEAN
#define NOGDI
#include <windows.h>
#include "nullsoft/in2.h"
#include "nullsoft/frontend.h"

#include "../libpcm/libpcm.h"
#include "../libpcm/fader.h"
#include "../pversion.h"

static const char winampplugin_copyright[] = WINAMPPLUGIN_COPYRIGHT;
static const char winampplugin_name[] = WINAMPPLUGIN_NAME;
static const char winampplugin_exts[] = WINAMPPLUGIN_EXTS;

#define SIZE_OF_BUFFER 4096

static struct
{
	HANDLE hThread; DWORD dwThreadId;
	HANDLE hEvent; volatile LONG request;
	int length_in_ms;
	int seek_pos;
	int paused;
	int samplerate; int samplebits; int ch; int blockalign;
	char title[256];
	unsigned char buffer[SIZE_OF_BUFFER];
} work;

static In_Module mod;

enum REQUEST
{
	REQUEST_STOP,
	REQUEST_SEEK
};

static void LIBPCM_CDECL init(void)
{
	work.hThread = 0;
	work.title[0] = '\0';
	work.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
}

static void LIBPCM_CDECL quit(void)
{
	if (work.hEvent)
	{
		CloseHandle(work.hEvent);
		work.hEvent = 0;
	}
}

static void LIBPCM_CDECL getfileinfo(char *fn, char *title, int *length_in_ms)
{
	LIBPCM_DECODER *d;
	LIBPCM_FADER *f;
	if (title) *title = '\0';
	if (length_in_ms) *length_in_ms = 0;
	if (!fn || !*fn)
	{
		lstrcpy(title, work.title);
		if (length_in_ms) *length_in_ms = work.length_in_ms;
		return;
	}
	d = libpcm_open_from_file(fn);
	if (d)
	{
		f = libpcm_fader_initialize(d);
		if (f)
		{
			if (title) lstrcpy(title, libpcm_get_title(d));
			if (length_in_ms) *length_in_ms = libpcm_fader_get_length_ms(f);
			libpcm_fader_terminate(f);
		}
	}
	libpcm_close(d);
}

static DWORD __stdcall PlayWorkerThread(void *argp)
{
	LIBPCM_FADER *f;
	LIBPCM_DECODER *d;
	unsigned cur, timeout, end_flag;
	int bufsize, len;

	int request;

	f = argp;
	d = libpcm_fader_get_decoder(f);

	work.paused = 0;
	work.request = 0;
	ResetEvent(work.hEvent);

	bufsize = SIZE_OF_BUFFER / work.blockalign;
	end_flag = cur = 0;
	do
	{
		cur = libpcm_get_currentposition_ms(d);
		len = libpcm_fader_read(f, work.buffer, bufsize);
		if (len != bufsize) end_flag = 1;
		if (len >= 576)
		{
			mod.SAAddPCMData(work.buffer, work.ch, work.samplebits, cur);
			mod.VSAAddPCMData(work.buffer, work.ch, work.samplebits, cur);
		}
		if(mod.dsp_isactive())
			len = mod.dsp_dosamples((short *)work.buffer, len, work.samplebits, work.ch, work.samplerate);
		len *= work.blockalign;
		timeout = 0;
		request = 0;
		do
		{
			if (WaitForSingleObject(work.hEvent, timeout) != WAIT_TIMEOUT)
			{
				request = InterlockedExchange((LPLONG)&work.request, 0);
				break;
			}
			timeout = 10;
		} while (mod.outMod->CanWrite() < len);
		if (request & (1 << REQUEST_STOP)) break;
		if (request & (1 << REQUEST_SEEK))
		{
			int seek_pos = work.seek_pos;
			mod.outMod->Flush(seek_pos);
			libpcm_seek_ms(d, seek_pos);
			work.seek_pos = 0;
			continue;
		}
		if (mod.outMod->CanWrite() >= len)
			mod.outMod->Write((char *)work.buffer, len);
	} while (!end_flag);

	if (end_flag)
	{
		timeout = 0;
		do
		{
			if (WaitForSingleObject(work.hEvent, timeout) != WAIT_TIMEOUT)
			{
				int request2 = InterlockedExchange((LPLONG)&work.request, 0);
				if (request2 & (1 << REQUEST_STOP))
				{
					request |= (1 << REQUEST_STOP);
					break;
				}
			}
			timeout = 10;
		} while (mod.outMod->IsPlaying());
	}
	libpcm_fader_terminate(f);
	libpcm_close(d);
	if (!(request & (1 << REQUEST_STOP)))
	{
		PostMessage(mod.hMainWindow, WM_WA_MPEG_EOF, 0, 0);
	}
	return 0;
}

static int LIBPCM_CDECL play(char *fn)
{
	int maxlatency;
	LIBPCM_DECODER *d;
	LIBPCM_FADER *f;

	if (!work.hEvent) return 1;

	d = libpcm_open_from_file(fn);
	if (!d) return -1;
	f = libpcm_fader_initialize(d);
	if (!f)
	{
		libpcm_close(d);
		return -1;
	}
	work.length_in_ms = libpcm_fader_get_length_ms(f);
	lstrcpyn(work.title, libpcm_get_title(d), 256);
	work.samplerate = libpcm_get_samplerate(d);
	work.samplebits = libpcm_get_bitspersample(d);
	work.ch = libpcm_get_numberofchannels(d);
	work.blockalign = libpcm_get_blockalign(d);
	maxlatency = mod.outMod->Open(work.samplerate, work.ch, work.samplebits, 0, 0);
	if (maxlatency < 0)
	{
		libpcm_fader_terminate(f);
		libpcm_close(d);
		return 1;
	}
	mod.SetInfo(libpcm_get_bitrate(d) / 1000, work.samplerate / 1000, work.ch, 0);
	mod.SAVSAInit(maxlatency, work.samplerate);
	mod.VSASetInfo(work.samplerate, work.ch);
	mod.outMod->SetVolume(-666);
	work.hThread = CreateThread(NULL, 0, PlayWorkerThread, f, 0, &work.dwThreadId);
	if (!work.hThread)
	{
		mod.SAVSADeInit();
		mod.outMod->Close();
		libpcm_fader_terminate(f);
		libpcm_close(d);
		return 1;
	}
	return 0;
}

static void Request(enum REQUEST request_code)
{
	if (work.hEvent)
	{
		work.request |= 1 << request_code;
		SetEvent(work.hEvent);
	}
}

static void LIBPCM_CDECL stop(void)
{
	if (work.hThread)
	{
		do
		{
			Request(REQUEST_STOP);
		} while (WaitForSingleObject(work.hThread, 20) == WAIT_TIMEOUT);
		CloseHandle(work.hThread);
		work.hThread = NULL;
		mod.SAVSADeInit();
		mod.outMod->Close();
	}
}

static void LIBPCM_CDECL pause(void) { work.paused = 1; mod.outMod->Pause(1); }
static void LIBPCM_CDECL unpause(void) { work.paused = 0; mod.outMod->Pause(0); }
static int LIBPCM_CDECL getoutputtime(void) { return mod.outMod->GetOutputTime(); }
static void LIBPCM_CDECL setvolume(int volume) { mod.outMod->SetVolume(volume); }
static void LIBPCM_CDECL setpan(int pan) { mod.outMod->SetPan(pan); }
static void LIBPCM_CDECL config(HWND hwnd) {}
static void LIBPCM_CDECL about(HWND hwnd) { MessageBoxA(hwnd, winampplugin_copyright , winampplugin_name, MB_OK); }
static int LIBPCM_CDECL infobox(char *fn, HWND hwnd) { return 0; }
static int LIBPCM_CDECL isourfile(char *fn) { return 0; }
static void LIBPCM_CDECL eq_set(int on, char data[10], int preamp) {}
static int LIBPCM_CDECL ispaused(void) { return work.paused; }
static int LIBPCM_CDECL getlength(void) { return work.length_in_ms; }
static void LIBPCM_CDECL setoutputtime(int time_in_ms)
{
	work.seek_pos = time_in_ms;
	Request(REQUEST_SEEK);
}

static const In_Module modinit = {
	IN_VER,
	(char *)winampplugin_name,
	0,	/* hMainWindow */
	0,	/* hDllInstance */
	(char *)winampplugin_exts,
	1,	/* is_seekable */
	1,	/* UsesOutputPlug */
	config,
	about,
	init,
	quit,
	getfileinfo,
	infobox,
	isourfile,
	play,
	pause,
	unpause,
	ispaused,
	stop,

	getlength,
	getoutputtime,
	setoutputtime,
	setvolume,
	setpan,

	/* pointers filled in by winamp */
	0,0,0,0,0,0,0,0,0,	/* vis stuff */
	0,0,	/* dsp */

	eq_set,	/* EQ, not used */
	NULL,	/* setinfo */
	0		/* out_mod */
};

In_Module * LIBPCM_CDECL winampGetInModule2(void)
{
	mod = modinit;
	return &mod;
}

#if defined(_MSC_VER) && !defined(_DEBUG)
#pragma comment(linker,"/OPT:NOWIN98")
#pragma comment(linker,"/ENTRY:DllMain")
#pragma comment(linker,"/MERGE:.rdata=.text")
#pragma comment(linker,"/STUB:../stub.exe")
#endif


BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		DisableThreadLibraryCalls(hinstDLL);
	}
#ifdef _DEBUG
	else if (fdwReason == DLL_PROCESS_DETACH)
	{
		extern int LIBPCM_API libpcm_leakcheck_dbg(void);
		if (libpcm_leakcheck_dbg())
		{
			MessageBoxA(0,"memman","leak",MB_OK);
		}
	}
#endif
	return TRUE;
}

