#include "studio/services/svc_fileread.h"

struct STREAM_READER : public libpcm_IReader
{
	volatile uint32_t refcount;
	svc_fileReader *reader;
	uint32_t size;
	const char *path;
};

static void LIBPCM_CALLBACK reader_seek(void *pv, uint32_t bytes)
{
	STREAM_READER *s = static_cast <STREAM_READER *> (pv);
	s->reader->seek(bytes);
}
static uint_t LIBPCM_CALLBACK reader_read(void *pv, void *buf, uint_t nbytes)
{
	STREAM_READER *s = static_cast <STREAM_READER *> (pv);
	return s->reader->read(static_cast<char *>(buf), nbytes);
}
static uint32_t LIBPCM_CALLBACK reader_size(void *pv)
{
	STREAM_READER *s = static_cast <STREAM_READER *> (pv);
	return s->size;
}
static void * LIBPCM_CALLBACK reader_addref(void *pv)
{
	STREAM_READER *s = static_cast <STREAM_READER *> (pv);
	s->refcount++;
	return s;
}
static void LIBPCM_CALLBACK reader_release(void *pv)
{
	STREAM_READER *s = static_cast <STREAM_READER *> (pv);
	if (--s->refcount == 0)
		delete s;
}
static const char * LIBPCM_CALLBACK reader_path(void *pv)
{
	STREAM_READER *s = static_cast <STREAM_READER *> (pv);
	return s->path;
}
static const libpcm_IReaderVtbl WacReaderVtbl =
{
	reader_addref,
	reader_release,
	reader_size,
	reader_read,
	reader_seek,
	reader_path,
};

libpcm_IReader * libpcm_CreateWacReader(svc_fileReader *preader, const char *path)
{
	if (!preader) return 0;
	if (preader->getLength() <= 0) return 0;
	//if (!preader->CanSeek()) return 0;
	STREAM_READER *s = new STREAM_READER;
	if (!s) return 0;
	s->lpVtbl = &WacReaderVtbl;
	s->refcount = 1;
	s->reader = preader;
	s->path = path;
	s->size = preader->getLength();
	return s;
}
